@@ -25,6 +25,7 @@
 
 #include <unistd.h>
 #include <fcntl.h>
+#include <sys/ioctl.h>
 #include <sys/select.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -32,6 +33,9 @@
 #include <sys/un.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#ifdef PSEUDO_EPOLL
+#include <sys/epoll.h>
+#endif
 #include <signal.h>
 
 #include "pseudo.h"
@@ -59,9 +63,11 @@ static int active_clients = 0, highest_client = 0, max_clients = 0;
 
 #define LOOP_DELAY 2
 #define DEFAULT_PSEUDO_SERVER_TIMEOUT 30
+#define EPOLL_MAX_EVENTS 10
 int pseudo_server_timeout = DEFAULT_PSEUDO_SERVER_TIMEOUT;
 static int die_peacefully = 0;
 static int die_forcefully = 0;
+static sig_atomic_t do_list_clients = 0;
 
 /* when the client is linked with pseudo_wrappers, these are defined there.
  * when it is linked with pseudo_server, though, we have to provide different
@@ -77,10 +83,19 @@ quit_now(int signal) {
 	die_forcefully = 1;
 }
 
+static void
+set_do_list_clients(int sig) {
+	do_list_clients = sig;
+}
+
 static int messages = 0, responses = 0;
 static struct timeval message_time = { .tv_sec = 0 };
 
+#ifdef PSEUDO_EPOLL
+static void pseudo_server_loop_epoll(void);
+#else
 static void pseudo_server_loop(void);
+#endif
 
 /* helper function to make a directory, just like mkdir -p.
  * Can't use system() because the child shell would end up trying
@@ -162,6 +177,9 @@ pseudo_server_start(int daemonize) {
 	 * SIGUSR1, or until too much time has passed. */
 	if (daemonize) {
 		int child;
+
+		/* make startup messages go away when invoked-as-daemon */
+		pseudo_debug_logfile(PSEUDO_LOGFILE, 2);
 		child = fork();
 		if (child == -1) {
 			pseudo_diag("Couldn't fork child process: %s\n",
@@ -231,7 +249,6 @@ pseudo_server_start(int daemonize) {
 			setsid();
 			fclose(stdin);
 			fclose(stdout);
-			pseudo_debug_logfile(PSEUDO_LOGFILE, 2);
 			/* and then just execute the server code normally.  */
 			/* Any logging will presumably go to logfile, but
 			 * exit status will make it back to the parent for
@@ -358,16 +375,25 @@ pseudo_server_start(int daemonize) {
 	signal(SIGTERM, quit_now);
 	/* tell parent process to stop waiting */
 	if (daemonize) {
-		pseudo_diag("Setup complete, sending SIGUSR1 to pid %d.\n",
-			getppid());
-		kill(getppid(), SIGUSR1);
+		pid_t ppid = getppid();
+		if (ppid == 1) {
+			pseudo_diag("Setup complete, but parent is init, not sending SIGUSR1.\n");
+		} else {
+			pseudo_diag("Setup complete, sending SIGUSR1 to pid %d.\n",
+				ppid);
+			kill(ppid, SIGUSR1);
+		}
 	}
+#ifdef PSEUDO_EPOLL
+	pseudo_server_loop_epoll();
+#else
 	pseudo_server_loop();
+#endif
 	return 0;
 }
 
 /* mess with internal tables as needed */
-static void
+static unsigned int
 open_client(int fd) {
 	pseudo_client_t *new_clients;
 	int i;
@@ -383,7 +409,7 @@ open_client(int fd) {
 			++active_clients;
 			if (i > highest_client)
 				highest_client = i;
-			return;
+			return i;
 		}
 	}
 
@@ -407,9 +433,11 @@ open_client(int fd) {
 
 		max_clients += 16;
 		++active_clients;
+		return max_clients - 16;
 	} else {
 		pseudo_diag("error allocating new client, fd %d\n", fd);
 		close(fd);
+		return 0;
 	}
 }
 
@@ -445,7 +473,10 @@ static int
 serve_client(int i) {
 	pseudo_msg_t *in;
 	int rc;
+	struct timeval tv1, tv2;
+	gettimeofday(&tv1, NULL);
 
+	++messages;
 	pseudo_debug(PDBGF_SERVER, "message from client %d [%d:%s - %s] fd %d\n",
 		i, (int) clients[i].pid,
 		clients[i].program ? clients[i].program : "???",
@@ -455,7 +486,7 @@ serve_client(int i) {
 	if (in) {
 		char *response_path = 0;
 		size_t response_pathlen;
-                int send_response = 1;
+		int send_response = 1;
 		pseudo_debug(PDBGF_SERVER | PDBGF_VERBOSE, "got a message (%d): %s\n", in->type, (in->pathlen ? in->path : "<no path>"));
 		/* handle incoming ping */
 		if (in->type == PSEUDO_MSG_PING && !clients[i].pid) {
@@ -482,6 +513,11 @@ serve_client(int i) {
 			}
 			pseudo_debug(PDBGF_SERVER, "\n");
 		}
+		if (in->type == PSEUDO_MSG_SHUTDOWN && !clients[i].pid) {
+			pseudo_debug(PDBGF_SERVER, "shutdown request from client %d [pid %d]\n",
+				i, in->client);
+			in->client = clients[i].pid;
+		}
 		/* sanity-check client ID */
 		if (in->client != clients[i].pid) {
 			pseudo_debug(PDBGF_SERVER, "uh-oh, expected pid %d for client %d, got %d\n",
@@ -491,8 +527,6 @@ serve_client(int i) {
 		 * pseudo_server_response.
 		 */
 		if (in->type != PSEUDO_MSG_SHUTDOWN) {
-                        if (in->type == PSEUDO_MSG_FASTOP)
-                                send_response = 0;
 			/* most messages don't need these, but xattr may */
 			response_path = 0;
 			response_pathlen = -1;
@@ -509,6 +543,20 @@ serve_client(int i) {
 			} else {
 				in->pathlen = 0;
 			}
+			if (in->type == PSEUDO_MSG_FASTOP) {
+				/* respond instantly */
+				send_response = 0;
+				int t_type = in->type;
+				int t_pathlen = in->pathlen;
+				in->type = PSEUDO_MSG_ACK;
+				in->pathlen = 0;
+				if ((rc = pseudo_msg_send(clients[i].fd, in, in->pathlen, response_path)) != 0) {
+					pseudo_debug(PDBGF_SERVER, "failed to send response to client %d [%d]: %d (%s)\n",
+						i, (int) clients[i].pid, rc, strerror(errno));
+				}
+				in->type = t_type;
+				in->pathlen = t_pathlen;
+			}
 		} else {
 			/* the server's listen fd is "a client", and
 			 * so is the program connecting to request a shutdown.
@@ -539,26 +587,203 @@ serve_client(int i) {
 				die_peacefully = 1;
 			}
 		}
-                if (send_response) {
-                        if ((rc = pseudo_msg_send(clients[i].fd, in, in->pathlen, response_path)) != 0) {
-                                pseudo_debug(PDBGF_SERVER, "failed to send response to client %d [%d]: %d (%s)\n",
-                                        i, (int) clients[i].pid, rc, strerror(errno));
-                        }
-                } else {
-                        rc = 1;
-                }
+		if (send_response) {
+			if ((rc = pseudo_msg_send(clients[i].fd, in, in->pathlen, response_path)) != 0) {
+				pseudo_debug(PDBGF_SERVER, "failed to send response to client %d [%d]: %d (%s)\n",
+					i, (int) clients[i].pid, rc, strerror(errno));
+			}
+		} else {
+			rc = 1;
+		}
 		free(response_path);
-		return rc;
 	} else {
 		/* this should not be happening, but the exceptions aren't
 		 * being detected in select() for some reason.
 		 */
 		pseudo_debug(PDBGF_SERVER, "client %d: no message\n", (int) clients[i].pid);
 		close_client(i);
-		return 0;
+		rc = 0;
 	}
+
+	gettimeofday(&tv2, NULL);
+	if (rc == 0)
+		++responses;
+	message_time.tv_sec += (tv2.tv_sec - tv1.tv_sec);
+	message_time.tv_usec += (tv2.tv_usec - tv1.tv_usec);
+	if (message_time.tv_usec < 0) {
+		message_time.tv_usec += 1000000;
+		--message_time.tv_sec;
+	} else while (message_time.tv_usec > 1000000) {
+		message_time.tv_usec -= 1000000;
+		++message_time.tv_sec;
+	}
+	return rc;
 }
 
+#ifdef PSEUDO_EPOLL
+static void pseudo_server_loop_epoll(void)
+{
+	struct sockaddr_un client;
+	socklen_t len;
+	int i;
+	int rc;
+	int fd;
+	int timeout;
+	struct epoll_event ev, events[EPOLL_MAX_EVENTS];
+	int loop_timeout = pseudo_server_timeout;
+	struct sigaction eat_usr2 = {
+		.sa_handler = set_do_list_clients
+	};
+
+	clients = malloc(16 * sizeof(*clients));
+
+	sigaction(SIGUSR2, &eat_usr2, NULL);
+
+	clients[0].fd = listen_fd;
+	clients[0].pid = getpid();
+
+	for (i = 1; i < 16; ++i) {
+		clients[i].fd = -1;
+		clients[i].pid = 0;
+		clients[i].tag = NULL;
+		clients[i].program = NULL;
+	}
+
+	active_clients = 1;
+	max_clients = 16;
+	highest_client = 0;
+
+	pseudo_debug(PDBGF_SERVER, "server loop started.\n");
+	if (listen_fd < 0) {
+		pseudo_diag("got into loop with no valid listen fd.\n");
+		exit(PSEUDO_EXIT_LISTEN_FD);
+	}
+
+	timeout = LOOP_DELAY * 1000;
+
+	int epollfd = epoll_create1(0);
+	if (epollfd == -1) {
+		pseudo_diag("epoll_create1() failed.\n");
+		exit(PSEUDO_EXIT_EPOLL_CREATE);
+	}
+	ev.events = EPOLLIN;
+	ev.data.u64 = 0;
+	if (epoll_ctl(epollfd, EPOLL_CTL_ADD, clients[0].fd, &ev) == -1) {
+		pseudo_diag("epoll_ctl() failed with listening socket.\n");
+		exit(PSEUDO_EXIT_EPOLL_CTL);
+	}
+
+	pdb_log_msg(SEVERITY_INFO, NULL, NULL, NULL, "server started (pid %d)", getpid());
+
+	for (;;) {
+		rc = epoll_wait(epollfd, events, EPOLL_MAX_EVENTS, timeout);
+		if (rc == 0 || (rc == -1 && errno == EINTR)) {
+			/* If there's no clients, start timing out.  If there
+			 * are active clients, never time out.
+			 */
+			if (active_clients == 1) {
+				loop_timeout -= LOOP_DELAY;
+				/* maybe flush database to disk */
+				pdb_maybe_backup();
+				if (loop_timeout <= 0) {
+					pseudo_debug(PDBGF_SERVER, "no more clients, got bored.\n");
+					die_peacefully = 1;
+				} else {
+					/* display this if not exiting */
+					pseudo_debug(PDBGF_SERVER | PDBGF_BENCHMARK, "%d messages handled in %.4f seconds, %d responses\n",
+						messages,
+						(double) message_time.tv_sec +
+						(double) message_time.tv_usec / 1000000.0,
+						responses);
+				}
+			}
+		} else if (rc > 0) {
+			loop_timeout = pseudo_server_timeout;
+			for (i = 0; i < rc; ++i) {
+				if (clients[events[i].data.u64].fd == listen_fd) {
+					if (!die_forcefully) {
+						len = sizeof(client);
+						if ((fd = accept(listen_fd, (struct sockaddr *) &client, &len)) != -1) {
+						/* Don't allow clients to end up on fd 2, because glibc's
+						 * malloc debug uses that fd unconditionally.
+						 */
+							if (fd == 2) {
+								int newfd = fcntl(fd, F_DUPFD, 3);
+								close(fd);
+								fd = newfd;
+							}
+							pseudo_debug(PDBGF_SERVER, "new client fd %d\n", fd);
+							/* A new client implicitly cancels any
+							 * previous shutdown request, or a
+							 * shutdown for lack of clients.
+							 */
+							pseudo_server_timeout = DEFAULT_PSEUDO_SERVER_TIMEOUT;
+							die_peacefully = 0;
+
+							ev.events = EPOLLIN;
+							ev.data.u64 = open_client(fd);
+							if (ev.data.u64 != 0 && epoll_ctl(epollfd, EPOLL_CTL_ADD, clients[ev.data.u64].fd, &ev) == -1) {
+								pseudo_diag("epoll_ctl() failed with accepted socket.\n");
+								exit(PSEUDO_EXIT_EPOLL_CTL);
+							}
+						} else if (errno == EMFILE) {
+							// select() loop would drop a client here, we do nothing (for now)
+							pseudo_debug(PDBGF_SERVER, "Hit max open files.\n");
+						}
+					}
+				} else {
+					int n = 0;
+					ioctl(clients[i].fd, FIONREAD, &n);
+					if (n == 0) {
+						close_client(i);
+					} else {
+						serve_client(i);
+					}
+				}
+				if (die_forcefully)
+					break;
+			}
+			pseudo_debug(PDBGF_SERVER, "server loop complete [%d clients left]\n", active_clients);
+		} else {
+			pseudo_diag("epoll_wait failed: %s\n", strerror(errno));
+			break;
+		}
+		if (do_list_clients) {
+			do_list_clients = 0;
+			pseudo_diag("listing clients [1 through %d]:\n", highest_client);
+			for (i = 1; i <= highest_client; ++i) {
+				if (clients[i].fd == -1) {
+					pseudo_diag("client %4d: inactive.\n", i);
+					continue;
+				}
+				pseudo_diag("client %4d: fd %4d, pid %5d, program %s\n",
+					i, clients[i].fd, clients[i].pid,
+					clients[i].program ? clients[i].program : "<unspecified>");
+			}
+			pseudo_diag("done.\n");
+		}
+		if (die_peacefully || die_forcefully) {
+			pseudo_debug(PDBGF_SERVER, "quitting.\n");
+			pseudo_debug(PDBGF_SERVER | PDBGF_BENCHMARK, "server %d exiting: handled %d messages in %.4f seconds\n",
+				getpid(), messages,
+				(double) message_time.tv_sec +
+				(double) message_time.tv_usec / 1000000.0);
+			pdb_log_msg(SEVERITY_INFO, NULL, NULL, NULL, "server %d exiting: handled %d messages in %.4f seconds",
+				getpid(), messages,
+				(double) message_time.tv_sec +
+				(double) message_time.tv_usec / 1000000.0);
+			/* and at this point, we'll start refusing connections */
+			close(clients[0].fd);
+			/* This is a good place to insert a delay for
+			 * debugging race conditions during startup. */
+			/* usleep(300000); */
+			exit(0);
+		}
+	}
+
+}
+#else
+
 /* get clients, handle messages, shut down.
  * This doesn't actually do any work, it just calls a ton of things which
  * do work.
@@ -574,9 +799,14 @@ pseudo_server_loop(void) {
 	int rc;
 	int fd;
 	int loop_timeout = pseudo_server_timeout;
+	struct sigaction eat_usr2 = {
+		.sa_handler = set_do_list_clients
+	};
 
 	clients = malloc(16 * sizeof(*clients));
 
+	sigaction(SIGUSR2, &eat_usr2, NULL);
+
 	clients[0].fd = listen_fd;
 	clients[0].pid = getpid();
 
@@ -599,8 +829,8 @@ pseudo_server_loop(void) {
 	pdb_log_msg(SEVERITY_INFO, NULL, NULL, NULL, "server started (pid %d)", getpid());
 
 	FD_ZERO(&reads);
-	FD_ZERO(&writes);
 	FD_ZERO(&events);
+	FD_ZERO(&writes);
 	FD_SET(clients[0].fd, &reads);
 	FD_SET(clients[0].fd, &events);
 	max_fd = clients[0].fd;
@@ -609,15 +839,20 @@ pseudo_server_loop(void) {
 	/* EINTR tends to come from profiling, so it is not a good reason to
 	 * exit; other signals are caught and set the flag causing a graceful
 	 * exit. */
+	sigset_t maskusr2;
+	sigemptyset(&maskusr2);
+	sigaddset(&maskusr2, SIGUSR2);
+	sigprocmask(SIG_BLOCK, &maskusr2, NULL);
 	while ((rc = select(max_fd + 1, &reads, &writes, &events, &timeout)) >= 0 || (errno == EINTR)) {
+		sigprocmask(SIG_UNBLOCK, &maskusr2, NULL);
 		if (rc == 0 || (rc == -1 && errno == EINTR)) {
 			/* If there's no clients, start timing out.  If there
 			 * are active clients, never time out.
 			 */
 			if (active_clients == 1) {
 				loop_timeout -= LOOP_DELAY;
-                                /* maybe flush database to disk */
-                                pdb_maybe_backup();
+				/* maybe flush database to disk */
+				pdb_maybe_backup();
 				if (loop_timeout <= 0) {
 					pseudo_debug(PDBGF_SERVER, "no more clients, got bored.\n");
 					die_peacefully = 1;
@@ -627,34 +862,21 @@ pseudo_server_loop(void) {
 						messages,
 						(double) message_time.tv_sec +
 						(double) message_time.tv_usec / 1000000.0,
-                                                responses);
+						responses);
 				}
 			}
 		} else if (rc > 0) {
 			loop_timeout = pseudo_server_timeout;
 			for (i = 1; i <= highest_client; ++i) {
-				if (clients[i].fd == -1)
+				if (clients[i].fd == -1) {
 					continue;
-				if (FD_ISSET(clients[i].fd, &events)) {
-					/* this should happen but doesn't... */
-					close_client(i);
 				} else if (FD_ISSET(clients[i].fd, &reads)) {
-					struct timeval tv1, tv2;
-                                        int rc;
-					gettimeofday(&tv1, NULL);
-					rc = serve_client(i);
-					gettimeofday(&tv2, NULL);
-					++messages;
-                                        if (rc == 0)
-                                                ++responses;
-					message_time.tv_sec += (tv2.tv_sec - tv1.tv_sec);
-					message_time.tv_usec += (tv2.tv_usec - tv1.tv_usec);
-					if (message_time.tv_usec < 0) {
-						message_time.tv_usec += 1000000;
-						--message_time.tv_sec;
-					} else while (message_time.tv_usec > 1000000) {
-						message_time.tv_usec -= 1000000;
-						++message_time.tv_sec;
+					int n = 0;
+					ioctl(clients[i].fd, FIONREAD, &n);
+					if (n == 0) {
+						close_client(i);
+					} else {
+						serve_client(i);
 					}
 				}
 				if (die_forcefully)
@@ -675,16 +897,31 @@ pseudo_server_loop(void) {
 					}
 					pseudo_debug(PDBGF_SERVER, "new client fd %d\n", fd);
 					open_client(fd);
-                                        /* A new client implicitly cancels any
-                                         * previous shutdown request, or a
-                                         * shutdown for lack of clients.
-                                         */
-                                        pseudo_server_timeout = DEFAULT_PSEUDO_SERVER_TIMEOUT;
-                                        die_peacefully = 0;
+					/* A new client implicitly cancels any
+					 * previous shutdown request, or a
+					 * shutdown for lack of clients.
+					 */
+					pseudo_server_timeout = DEFAULT_PSEUDO_SERVER_TIMEOUT;
+					die_peacefully = 0;
 				}
 			}
 			pseudo_debug(PDBGF_SERVER, "server loop complete [%d clients left]\n", active_clients);
 		}
+		if (do_list_clients) {
+			do_list_clients = 0;
+			pseudo_diag("listing clients [1 through %d]:\n", highest_client);
+			for (i = 1; i <= highest_client; ++i) {
+				if (clients[i].fd == -1) {
+					pseudo_diag("client %4d: inactive.\n", i);
+					continue;
+				}
+				pseudo_diag("client %4d: fd %4d, pid %5d, state %s, program %s\n",
+					i, clients[i].fd, clients[i].pid,
+					FD_ISSET(clients[i].fd, &reads) ? "R" : "-",
+					clients[i].program ? clients[i].program : "<unspecified>");
+			}
+			pseudo_diag("done.\n");
+		}
 		if (die_peacefully || die_forcefully) {
 			pseudo_debug(PDBGF_SERVER, "quitting.\n");
 			pseudo_debug(PDBGF_SERVER | PDBGF_BENCHMARK, "server %d exiting: handled %d messages in %.4f seconds\n",
@@ -717,7 +954,6 @@ pseudo_server_loop(void) {
 			if (clients[i].fd != -1) {
 				++current_clients;
 				FD_SET(clients[i].fd, &reads);
-				FD_SET(clients[i].fd, &events);
 				if (clients[i].fd > max_fd)
 					max_fd = clients[i].fd;
 			}
@@ -728,6 +964,8 @@ pseudo_server_loop(void) {
 		}
 		/* reinitialize timeout because Linux select alters it */
 		timeout = (struct timeval) { .tv_sec = LOOP_DELAY, .tv_usec = 0 };
+		sigprocmask(SIG_BLOCK, &maskusr2, NULL);
 	}
 	pseudo_diag("select failed: %s\n", strerror(errno));
 }
+#endif /* this is the else of #ifdef PSEUDO_EPOLL */
